from shapely.wkt import loads, dumps
from pyproj import Proj, transform as pyproj_transform
from math import log10
from json import loads as json_loads
from shapely.geometry import shape, box
from functools import partial
from shapely.ops import transform as shapely_transform

from shapely.geometry import (
    Polygon, 
    Point, 
    MultiPolygon, 
    LineString,
    LinearRing,
    MultiLineString, 
    MultiPoint, 
    GeometryCollection
)

BASE32 = '0123456789bcdefghjkmnpqrstuvwxyz'

VALID_WKT_TYPES = [
    'GEOMETRY', 
    'POINT', 
    'MULTIPOINT', 
    'LINESTRING',
    'LINEARRING'
    'MULTILINESTRING', 
    'POLYGON', 
    'MULTIPOLYGON'
]


VALID_SHAPELY_TYPES = (
    Polygon, 
    Point, 
    MultiPolygon, 
    LineString,
    LinearRing,
    MultiLineString, 
    MultiPoint, 
    GeometryCollection
)

webmercator_project = partial(
    pyproj_transform,
    Proj(init='epsg:4326'),  # source coordinate system
    Proj(init='epsg:3857')  # destination coordinate system
)


def to_webmercator(pol):
    return shapely_transform(webmercator_project, pol)


def divide_range_decode(coordinate_range, b):
    mid = (coordinate_range[0] + coordinate_range[1]) / 2
    if b:
        coordinate_range[0] = mid
    else:
        coordinate_range[1] = mid


def geohash_to_centroid(geohash, return_error=True):
    binary_string = ''.join([format(BASE32.index(v), 'b').zfill(5) for v in geohash])
    latitude_range = [-90.0, 90.0]
    longitude_range = [-180.0, 180.0]
    is_even_bit = True

    for j, v in enumerate(binary_string):
        if is_even_bit:
            divide_range_decode(longitude_range, v != '0')
        else:
            divide_range_decode(latitude_range, v != '0')
        is_even_bit = not is_even_bit

    latitude = (latitude_range[0] + latitude_range[1]) / 2
    longitude = (longitude_range[0] + longitude_range[1]) / 2

    if return_error:
        latitude_error = abs(latitude_range[0] - latitude_range[1]) / 2
        longitude_error = abs(longitude_range[0] - longitude_range[1]) / 2
        return latitude, longitude, latitude_error, longitude_error
    else:
        return latitude, longitude


def validate_geohash(value):
    """Check whether an input value is a valid geohash"""
        
    if len(value) > 12:
        return False
    
    return all(v in BASE32 for v in value)


def detect_geojson(value):
    """Check whether an input value is a valid geohash"""

    if isinstance(value, str):
        try:
            _ = json_loads(value)
            return True
        except Exception:
            return False
    return False


def import_geojson(feature, precision=6):

    json_dict = json_loads(feature)
    if 'geometry' in json_dict:
        shape_object = shape(json_dict["geometry"])
    elif 'coordinates' in json_dict:
        shape_object = shape(json_dict)
    else:
        raise KeyError('Unable to infer key for coordinate values.')

    if isinstance(shape_object, Point):
        shape_object = shape_object.buffer(0.1 ** precision).envelope

    return shape_object


def validate_wellknowntext(value):
    """Check whether an input value is valid well-known text"""
        
    return any(value.startswith(vt) for vt in VALID_WKT_TYPES)

    
def validate_shapelyobject(value):
    """Check whether an input value is a valid shapely object."""
     
    return issubclass(type(value), VALID_SHAPELY_TYPES)
    

def geohash_to_shape(value):
    """Convert a geohash to a shapely object."""

    y, x, y_margin, x_margin = geohash_to_centroid(value)

    minx = x - x_margin
    maxx = x + x_margin
    miny = y - y_margin
    maxy = y + y_margin

    return box(minx, miny, maxx, maxy)


def geohash_to_coords(value, precision=6):
    """Convert a geohash to the x and y values of that geohash's bounding box."""
    y, x, y_margin, x_margin = geohash_to_centroid(value)
        
    x_coords = [
        round(x - x_margin, precision),
        round(x - x_margin, precision),
        round(x + x_margin, precision),
        round(x + x_margin, precision)
    ]
        
    y_coords = [
        round(y + y_margin, precision),
        round(y - y_margin, precision),
        round(y - y_margin, precision),
        round(y + y_margin, precision)
    ]

    return [{'exterior': x_coords, 'holes': []}], [{'exterior': y_coords, 'holes': []}]


def shape_to_coords(value, precision=6, wkt=False, is_point=False):
    """
    Convert a shape (a shapely object or well-known text) to x and y coordinates
    suitable for use in Bokeh's `MultiPolygons` glyph.
    
    """
    
    if is_point:
        value = Point(*value).buffer(0.1 ** precision).envelope
    
    x_coords = list()
    y_coords = list()
        
    if wkt:
        value = loads(value)
    if not hasattr(value, '__len__'):
        value = [value] 
        
    for v in value:
        x_dict = dict()
        y_dict = dict()
        if not hasattr(v, 'exterior'):
            v = v.buffer(0)
        x_dict['exterior'] = [round(x, precision) for x in v.exterior.coords.xy[0]]
        x_dict['holes'] = [[round(y, precision) for y in x.coords.xy[0]] for x in v.interiors]
        y_dict['exterior'] = [round(x, precision) for x in v.exterior.coords.xy[1]]
        y_dict['holes'] = [[round(y, precision) for y in x.coords.xy[1]] for x in v.interiors]
        x_coords.append(x_dict)
        y_coords.append(y_dict)

    return x_coords, y_coords


def transform_lon(v, precision):
    v = pyproj_transform(Proj(init='epsg:4326'), Proj(init='epsg:3857'), v, 0)[0]
    return round(v, precision)


def transform_lat(v, precision):
    v = pyproj_transform(Proj(init='epsg:4326'), Proj(init='epsg:3857'), 0, v)[1]
    return round(v, precision)


def transform(v, precision=6, longitude=True):
    return transform_lon(v, precision) if longitude else transform_lat(v, precision)


def coord_to_webmercator(c, precision=6, longitude=True):

    p = transform(0.1 ** precision, precision, longitude)
    p = int(round(log10(p))) * -1

    if not hasattr(c, '__len__'):
        return transform(c, p, longitude)

    output = list()
    for row in c:
        if not hasattr(row, '__len__'):
            row = transform(row, p, longitude)
            output.append(row)
        else:
            newrow = list()
            for v in row:
                v_new = dict()
                v_new['exterior'] = [transform(val, p, longitude) for val in v['exterior']]
                v_new['holes'] = [
                    [transform(val, p, longitude) for val in interior] 
                    for interior in v['holes']
                ]
                newrow.append(v_new)

            output.append(newrow)

    return output


def validate_latlon_pair(value):
    """
    Check to make sure value is a tuple or list of length 2,
    where the first element is a longitude and the second.
    is a latitude.
    """

    if len(value) != 2:
        raise ValueError('List inputs are assumed to be coordinate pairs. This list has more than two elements.')

    if not (-180.0 <= value[0] <= 180.0):
        raise ValueError(
            ' '.join([
                'The first element of the list is assumed to be longitude (x-coordinates),',
                'but this element is outside acceptable bounds (-180.0, 180.0)'
            ])
        )

    if not (-90.0 <= value[0] <= 90.0):
        raise ValueError(
            ' '.join([
                'The second element of the list is assumed to be latitude (y-coordinates),',
                'but this element is outside acceptable bounds (-90.0, 90.0)'
            ])
        )

    return True


def process_input_value(value):
    """
    Router function for values: take an arbitrary value,
    determine what kind of input it is, and return
    a standardized coordinate output.

    """

    if type(value) == str:
        if validate_geohash(value):
            return geohash_to_shape(value)
        elif validate_wellknowntext(value):
            return loads(value)
        else:
            raise ValueError('String inputs must be either a geohash or well-known text.')
    elif type(value) in (tuple, list):
        if validate_latlon_pair(value):
            return Point(*value)
    elif validate_shapelyobject(value):
        return value
    else:
        raise ValueError('Unrecognizeable input: {}.'.format(str(value)))


def polygon_to_nested_list(pol):

    ext_x, ext_y = zip(*list(pol.exterior.coords))
    x, y = [list(ext_x)], [list(ext_y)]
    for interior in pol.interiors:
        int_x, int_y = zip(*interior.coords)
        x.append(list(int_x))
        y.append(list(int_y))
    return x, y


def shape_to_nested_list(shp, buffer=0.000001):
    if shp.geometryType() in ('LineString', 'LinearRing'):
        xs, ys = polygon_to_nested_list(shp.buffer(buffer))
        xs, ys = [xs], [ys]
    elif shp.geometryType() == 'Point':
        xs, ys = polygon_to_nested_list(shp.buffer(buffer).envelope)
        xs, ys = [xs], [ys]
    elif shp.geometryType() == 'Polygon':
        xs, ys = polygon_to_nested_list(shp)
        xs, ys = [xs], [ys]
    elif shp.geometryType().startswith('Multi') or (shp.geometryType() == 'GeometryCollection'):
        xs, ys = list(zip(*[shape_to_nested_list(pol) for pol in shp]))
        xs, ys = [x[0] for x in xs], [y[0] for y in ys]
    else:
        raise NotImplementedError('Unrecognized shapely shape type.')
    return xs, ys


def dumps_if_shapely(ob):
    if validate_shapelyobject(ob):
        return dumps(ob)
    return ob



from bokeh.io import output_notebook, show, save
from bokeh.models.glyphs import Quadratic, Segment
from bokeh.models import GMapPlot, GMapOptions, ColumnDataSource, Range1d, Plot
from bokeh.models.tools import HoverTool, WheelZoomTool, ResetTool, PanTool, TapTool
from bokeh.models.annotations import Title, Legend, LegendItem
from bokeh.resources import CDN
from bokeh.models.layouts import Row, Column, WidgetBox
from bokeh.models import CustomJS, CustomJSFilter, CDSView, OpenURL
from bokeh.models import DataTable, TableColumn
from bokeh.models import LinearAxis, MercatorTicker, MercatorTickFormatter

from os import path
from pandas import DataFrame
from numpy import array

from . import bokeh_utils, coordinate_utils, gmaps_utils


class WorkflowOrderError(Exception):
    pass


def flatten(items, seqtypes=(list, tuple)):
    for i, x in enumerate(items):
        while i < len(items) and isinstance(items[i], seqtypes):
            items[i:i+1] = items[i]
    return items


class Theto(object):
    """
    This class provides a wrapper to produce most of the boilerplate needed to use Bokeh to plot on 
    top of Google Maps or other tile images.
    
    Example usage:
    
    geohashes = [
        'dnrgrfm', 'dnrgrf3', 'dnrgrf7', 'dnrgrf5', 'dnrgrfk', 'dnrgrf1', 
        'dnrgrfh', 'dnrgrf2', 'dnrgrcu', 'dnrgrfj', 'dnrgrcv', 'dnrgrf4'
    ]
    
    API_KEY = (GET API KEY AT developers.google.com/maps/documentation/javascript/get-api-key)
    
    (
        Theto(api_key=API_KEY)
        .add_source(geohashes, label='a', order=range(len(geohashes)))
        .prepare_plot(plot_width=700)
        .add_layer('a', Patches, color='yellow', alpha=0.5)
        .add_layer('a', Circle, color='blue', size=20, alpha=0.75)
        .add_layer(
            'a', Text, text='order', text_color='white', text_font_size='10pt', text_align='center',
            text_baseline='middle'
        )
        .render_plot('auto')
    )
    """
    
    def __init__(self, api_key=None, precision=6, autohide=True, padding=0.05):
        """
        Instantiate the class.
        
        Parameters
        
        api_key (str): A Google Maps API key, which can be obtained at
            developers.google.com/maps/documentation/javascript/get-api-key.
            If None, only non-Google tile providers ('osm', 'esri', 
            'wikipedia', 'cartodb') will be available.

        precision (int): The number of decimals places that are meaningful for
            the plot. Defaults to 6, which is about as precise as phone's
            GPS signal can be expected to be. This precision will get automatically
            transformed in cases where latitude and longitude are projected
            to web Mercator maps.

        autohide (bool): if True, the toolbar on plots will be hidden unless the
            mouse is hovering over the plot itself.

        padding (float): the minimum amount of space that should exist between the
            plotted objects and the edge of the map, expressed as a percentage
            of the range of the relevant direction. So a padding factor of 0.05 (the
            default) will multiply the factor by the maximum latitude minus the minimum
            latitude and add that amount to the top and bottom edges of the graph, and
            then multiply the factor by the maximum longitude minus the minimum
            longitdue and add that amount to the right and left edges of the graph.
        
        """
        self.api_key = api_key
        self.precision = precision
        self.autohide = autohide
        self.padding = padding
        self.colorbar = None

        # removed 'x_coord_point', 'y_coord_point', 'raw_data'
        self.omit_columns = [
            'index', 'x_coords', 'y_coords', 'x_coords_transform', 'y_coords_transform',
            'x_coord_point_transform', 'y_coord_point_transform'
        ]

        self.sources = dict()
        self.columndatasources = dict()
        self.views = dict()
        self.widgets = dict()
        self.data_tables = list()
        self.custom_js = None
        self.xmin = None
        self.ymin = None
        self.xmax = None
        self.ymax = None
        self.plot = None
        self.legend = Legend(location='bottom_center', click_policy='hide', background_fill_color='#fafafa')
        self.validation = {
            'add_source': False,
            'add_widget': False,
            'prepare_plot': False,
            'add_layer': False,
            'render_plot': False
        }
        self.remove_columns = dict()
        
    def _validate_workflow(self, stage):
        """
        Checks the input stage against previously executed stages and raises
        an error if things have been done out of order.
        
        """
        
        if stage == 'add_source':
            if self.validation['add_widget']:
                raise WorkflowOrderError('New sources cannot be added after `add_widget` has been called.')
            if self.validation['prepare_plot']:
                raise WorkflowOrderError('New sources cannot be added after `prepare_plot` has been called.')
            if self.validation['add_layer']:
                raise WorkflowOrderError('New sources cannot be added after `add_layer` has been called.')
            if self.validation['render_plot']:
                raise WorkflowOrderError('Method `render_plot` has already been called. Start new workflow.')
        if stage == 'add_widget':
            if not self.validation['add_source']:
                raise WorkflowOrderError('Method `add_source` must be called before adding a widget.')
            if self.validation['prepare_plot']:
                raise WorkflowOrderError('New widgets cannot be added after `prepare_plot` has been called.')
            if self.validation['add_layer']:
                raise WorkflowOrderError('New widgets cannot be added after `add_layer` has been called.')
            if self.validation['render_plot']:
                raise WorkflowOrderError('Method `render_plot` has already been called. Start new workflow.')
        if stage == 'prepare_plot':
            if not self.validation['add_source']:
                raise WorkflowOrderError('Method `add_source` must be called before calling `prepare_plot`.')
            if self.validation['add_layer']:
                raise WorkflowOrderError('Plot cannot be prepared after `add_layer` has been called.')
            if self.validation['render_plot']:
                raise WorkflowOrderError('Method `render_plot` has already been called. Start new workflow.')
        if stage == 'add_layer':
            if not self.validation['add_source']:
                raise WorkflowOrderError('Method `add_source` must be called before adding a layer.')
            if not self.validation['prepare_plot']:
                raise WorkflowOrderError('New layers cannot be added after `prepare_plot` has been called.')
            if self.validation['render_plot']:
                raise WorkflowOrderError('Method `render_plot` has already been called. Start new workflow.')
        if stage == 'render_plot':
            if not self.validation['add_source']:
                raise WorkflowOrderError('Method `add_source` must be called before rendering the plot.')
            if not self.validation['prepare_plot']:
                raise WorkflowOrderError('Method `prepare_plot` must be called before rendering the plot.')
            if not self.validation['add_layer']:
                raise WorkflowOrderError('Method `add_layer` must be called before rendering the plot.')
        if stage == 'add_path':
            if not self.validation['add_source']:
                raise WorkflowOrderError('Method `add_source` must be called before adding a path.')
            if not self.validation['prepare_plot']:
                raise WorkflowOrderError('Method `prepare_plot` must be called before adding a path.')
            if self.validation['render_plot']:
                raise WorkflowOrderError('Method `render_plot` has already been called. Start new workflow.')
            if self.validation['add_widget']:
                raise WorkflowOrderError('Plot contains widgets, which are not compatible with paths.')
        if stage == 'add_data_table':
            if not self.validation['add_source']:
                raise WorkflowOrderError('Method `add_source` must be called before adding a data table.')
            if not self.validation['prepare_plot']:
                raise WorkflowOrderError('Method `prepare_plot` must be called before adding a data table.')
            if self.validation['render_plot']:
                raise WorkflowOrderError('Method `render_plot` has already been called. Start new workflow.')

    def _set_coordinate_bounds(self, df):
        """
        Given a new source DataFrame, set or update 
        coordinate bounds for the plot.
        
        """

        x_coords = flatten(df['x_coords'].tolist())
        y_coords = flatten(df['y_coords'].tolist())

        xmin = min(x_coords)
        xmax = max(x_coords)
        ymin = min(y_coords)
        ymax = max(y_coords)
        
        if self.xmin is not None:
            if xmin < self.xmin:
                self.xmin = xmin
        else:
            self.xmin = xmin

        if self.xmax is not None:
            if xmax > self.xmax:
                self.xmax = xmax
        else: 
            self.xmax = xmax

        if self.ymin is not None:
            if ymin < self.ymin:
                self.ymin = ymin
        else:
            self.ymin = ymin

        if self.ymax is not None:
            if ymax > self.ymax:
                self.ymax = ymax
        else:
            self.ymax = ymax

    def add_source(self, data, label, uid=None, column_name=None, **kwargs):
        """
        Add a source to the `self.sources`, to be used in the plot.
        
        Parameters:
        
        data (list or DataFrame): a list of geohashes, shapely objects, well-known text 
            strings, or longitude/latitude pairs; or a DataFrame where the objects to be 
            plotted are indicated by `column_name`
        label (str): a name that can be used to reference the source going forward
        uid (str or list): a list of unique ids (necessary for pathing); if None, `range` 
            is used to assign a number. If None, either `data` (if data is a list) or 
            `data[column_name]` (if data is a DataFrame) will be used.
        column_name (str): optional, the column of `data` that contains geohashes, shapely 
            objects, well-known text strings, or longitude/latitude pairs
        kwargs: lists of the same length as `data` if `data` is a list. These will be 
            appended to the data as metadata
        
        """
        
        self._validate_workflow('add_source')

        # process data into x and y coordinates
        if type(data) == DataFrame:
            if column_name is None:
                raise ValueError('If data is a dataframe then column_name must be specified.')
            df = data.copy()
            raw_data = df[column_name].values.tolist()
        else:
            raw_data = list(data)
            df = DataFrame(index=range(len(raw_data)))
            column_name = 'raw_data'

        if all(coordinate_utils.detect_geojson(v) for v in raw_data):
            processed_data = [coordinate_utils.import_geojson(v, self.precision) for v in raw_data]
        else:
            processed_data = [coordinate_utils.process_input_value(v) for v in raw_data]

        processed_data_transformed = [coordinate_utils.to_webmercator(v) for v in processed_data]
        buf = 1 / (10 ** self.precision)

        # original projection
        x_coords_shape, y_coords_shape = zip(*[coordinate_utils.shape_to_nested_list(v, buf) for v in processed_data])
        df['x_coords'], df['y_coords'] = list(x_coords_shape), list(y_coords_shape)

        x_coords_point, y_coords_point = zip(*[list(*v.representative_point().coords) for v in processed_data])
        df['x_coord_point'] = [round(v, self.precision) for v in x_coords_point]
        df['y_coord_point'] = [round(v, self.precision) for v in y_coords_point]

        # webmercator projection
        x_coords_shape, y_coords_shape = zip(*[
            coordinate_utils.shape_to_nested_list(v, buf) for v in processed_data_transformed
        ])
        df['x_coords_transform'], df['y_coords_transform'] = list(x_coords_shape), list(y_coords_shape)

        x_coords_point, y_coords_point = zip(*[
            list(*v.representative_point().coords) for v in processed_data_transformed
        ])
        df['x_coord_point_transform'] = [round(v, self.precision) for v in x_coords_point]
        df['y_coord_point_transform'] = [round(v, self.precision) for v in y_coords_point]

        if len(kwargs) > 0:
            for k, v in kwargs.items():
                df[k] = v

        self._set_coordinate_bounds(df)

        df[column_name] = [coordinate_utils.dumps_if_shapely(ob) for ob in raw_data]

        if uid is None:
            df['uid'] = range(df.shape[0])
        elif type(uid) is str:
            df['uid'] = df[uid]
        else:
            df['uid'] = uid
            
        self.sources[label] = df.copy()
        self.remove_columns[label] = ['f', 'p']
        self.validation['add_source'] = True
        
        return self

    def add_widget(self, source_label, widget_type, reference, widget_name=None, custom_js=None, **kwargs):
        """
        Add widgets to subset the data displayed in the plot.
        
        Parameters:
            source_label (str): the label for the data source to be used for the widget
            widget_type (str): a widget that is a valid key from
                `bokeh_utils.WIDGETS`.
            widget_name (str): an arbitrary name that can be used to identify the widget
                (necessary if `custom_js` is not None).
            reference (str): a column name from `self.sources[source_label]` that the widget
                should look at to filter the data source
            custom_js (str): custom javascript code for defining the filtering fuunction
            kwargs: all keyword arguments will be fed the instantiation of the widget to define
                the look and content of the interaction (ranges, initial values, etc.)
        
        """

        self._validate_workflow('add_widget')

        if widget_type == 'Animation':
            widget_type = 'Slider'
            animate = True
        else:
            animate = False

        if widget_name is None:
            widget_name = widget_type
        
        if reference not in self.sources[source_label].columns:
            raise ValueError("Reference '{}' not in data source '{}'.".format(reference, source_label))
                
        source = self.sources[source_label]
        
        ref_array = source[reference].tolist()

        animation_kwargs = {'button_type': 'primary', 'label': 'Start', 'ms_delay': 500}
        remove_kw = list()

        for k, v in kwargs.items():
            if k.startswith('animation'):
                remove_kw.append(k)
                k = k.replace('animation_', '')
                animation_kwargs[k] = v

        for k in remove_kw:
            _ = kwargs.pop(k)

        for k, v in bokeh_utils.DEFAULT_KWARGS[widget_type].items():
            if k not in kwargs:
                kwargs[k] = v

        widget = bokeh_utils.WIDGETS[widget_type](name=widget_name, **{
            k: v if v != 'auto' else bokeh_utils.auto_widget_kwarg(widget_type, k, ref_array)
            for k, v in kwargs.items()
        })

        # This callback is crucial to trigger changes when the widget changes
        callback = CustomJS(args=dict(source=None), code="""
            source.change.emit();
            console.log()
        """)
        use_active = widget_type in ('CheckboxButtonGroup', 'CheckboxGroup', 'RadioButtonGroup')
        widget.js_on_change('active' if use_active else 'value', callback)
        
        if custom_js is None:
            if self.custom_js is not None:
                raise ValueError('Either all widgets or no widgets must use custom_js.')
                
            if source_label in self.views:
                raise ValueError('Multiple widgets can only be used with custom_js.')
                
            js_filter = CustomJSFilter(
                args=dict(widget=widget, reference=reference), 
                code=bokeh_utils.FILTERS[widget_type]
            )
            
        else:
            if self.custom_js is None:
                self.custom_js = CustomJSFilter(args={}, code=custom_js)
                
            js_filter = self.custom_js
            js_filter.args.update({widget_name: widget, widget_name + '_ref': reference})

        self.widgets[widget_name] = {'widget': widget, 'filter': js_filter, 'source': source_label}

        if animate:
            ms_delay = animation_kwargs.pop('ms_delay')
            button = bokeh_utils.Button(**animation_kwargs)
            button.js_on_event(bokeh_utils.ButtonClick, bokeh_utils.auto_advance(widget, ms_delay))
            self.widgets['animation'] = {'widget': button, 'filter': None, 'source': None}

        if source_label not in self.views:
            self.views[source_label] = CDSView(filters=[js_filter])
            
        self.validation['add_widget'] = True
        
        return self

    def _create_columndatasource(self, source_df):
        """
        Convert a source DataFrame into a Bokeh ColumnDataSource
        based on what type of plot has been chosen.
        
        """
        
        suffix = '_transform' if type(self.plot) != GMapPlot else ''
        x_coords_label = 'x_coords{}'.format(suffix)
        y_coords_label = 'y_coords{}'.format(suffix)
        x_point_label = 'x_coord_point{}'.format(suffix)
        y_point_label = 'y_coord_point{}'.format(suffix)

        xsf = source_df[x_coords_label].tolist()
        ysf = source_df[y_coords_label].tolist()

        xsp = source_df[x_point_label].tolist()
        ysp = source_df[y_point_label].tolist()
        
        omit = [c for c in self.omit_columns if c in source_df.columns]
        source = ColumnDataSource(source_df.drop(omit, axis=1).to_dict('list'))
        source.data['xsf'] = xsf
        source.data['ysf'] = ysf
        source.data['xsp'] = xsp
        source.data['ysp'] = ysp
        
        return source
    
    def prepare_plot(
        self, plot_width=700, plot_height=None, zoom=None, map_type='carto_light',
        title=None, **kwargs
    ):
        """
        Create the actual plot object (stored in `self.plot`).
        
        Parameters:
        
        plot_width (int): desired plot width, in pixels
        plot_height (int): desired plot height, will be calculated 
            automatically if not supplied
        zoom (int): zoom factor for Google Maps, will be calculated 
            automatically if not supplied
        map_type (string): 'satellite', 'roadmap', or 'hybrid'
        title (string or tuple): if string, title is added to plot; 
            if tuple, the first value is the title, and second value 
            is a dict of kwargs
        kwargs: any options passed to Bokeh GMapPlot (title, etc.)
        
        """
        
        self._validate_workflow('prepare_plot')
        
        zoom_level, lat_center, lng_center, auto_plot_height = gmaps_utils.estimate_zoom(
            plot_width,
            x_bounds=(self.xmin, self.xmax),
            y_bounds=(self.ymin, self.ymax)
        )
        if plot_height is None:
            plot_height = auto_plot_height
            
        if zoom is None:
            zoom = zoom_level
        
        if title is not None:
            if isinstance(title, str):
                title = Title(text=title)
            if isinstance(title, (list, tuple)):
                title, title_kwargs = title
                title = Title(text=title, **title_kwargs)
                
        if map_type in ('satellite', 'roadmap', 'terrain', 'hybrid'):
            if self.api_key is None:
                raise ValueError(
                    'Class must be instantiated with Google Maps API key to use map_type `{}`'.format(map_type)
                )

            map_options = GMapOptions(lat=lat_center, lng=lng_center, map_type=map_type, zoom=zoom)
            self.plot = GMapPlot(
                x_range=Range1d(),
                y_range=Range1d(),
                map_options=map_options,
                plot_width=plot_width,
                plot_height=plot_height,
                title=title,
                **kwargs
            )
            self.plot.api_key = self.api_key
            self.plot.add_tools(WheelZoomTool(), ResetTool(), PanTool(), TapTool())
        elif map_type in bokeh_utils.get_tile_source(None):
            x_rng, y_rng = self.xmax - self.xmin, self.ymax - self.ymin
            x_range = Range1d(
                start=coordinate_utils.coord_to_webmercator(
                    self.xmin - (x_rng * self.padding),
                    precision=self.precision, 
                    longitude=True
                ), 
                end=coordinate_utils.coord_to_webmercator(
                    self.xmax + (x_rng * self.padding),
                    precision=self.precision,
                    longitude=True
                )
            )
            y_range = Range1d(
                start=coordinate_utils.coord_to_webmercator(
                    self.ymin - (y_rng * self.padding),
                    precision=self.precision,
                    longitude=False
                ), 
                end=coordinate_utils.coord_to_webmercator(
                    self.ymax + (y_rng * self.padding),
                    precision=self.precision,
                    longitude=False
                )
            )
            self.plot = Plot(
                x_range=x_range,
                y_range=y_range,
                frame_width=plot_width,
                frame_height=plot_height,
                title=title,
                **kwargs
            )
            
            self.plot.add_tile(bokeh_utils.get_tile_source(map_type))
            self.plot.add_tools(WheelZoomTool(), ResetTool(), PanTool(), TapTool())

            xformatter = MercatorTickFormatter(dimension="lon")
            xticker = MercatorTicker(dimension="lon")
            xaxis = LinearAxis(
                formatter=xformatter, ticker=xticker,
                axis_line_alpha=0.1, minor_tick_line_alpha=0.1, major_tick_line_alpha=0.1, major_label_text_alpha=0.5
            )
            self.plot.add_layout(xaxis, 'below')

            yformatter = MercatorTickFormatter(dimension="lat")
            yticker = MercatorTicker(dimension="lat")
            yaxis = LinearAxis(
                formatter=yformatter, ticker=yticker,
                axis_line_alpha=0.1, minor_tick_line_alpha=0.1, major_tick_line_alpha=0.1, major_label_text_alpha=0.5
            )
            self.plot.add_layout(yaxis, 'left')

        else:
            raise ValueError('Invalid map_type.')
                    
        for source_label, source in self.sources.items():
            source = self._create_columndatasource(source)
            
            for widget_name in self.widgets.keys():
                widget_dict = self.widgets[widget_name]
                if widget_dict['source'] == source_label:
                    widget_dict['filter'].args['source'] = source
                    for callback_list in widget_dict['widget'].js_property_callbacks.values():
                        for callback in callback_list:
                            callback.args['source'] = source
                    
            if source_label in self.views:
                self.views[source_label].source = source
                
            self.columndatasources[source_label] = source
            
        self.validation['prepare_plot'] = True
        
        return self
        
    def add_layer(
        self, source_label, bokeh_model='MultiPolygons', tooltips=None, legend=None, click_for_map=None,
        start_hex='#ff0000', end_hex='#0000ff', mid_hex='#ffffff', color_transform=None,
        **kwargs
    ):
        """
        Add bokeh models (glyphs or markers) to `self.plot`. 
        `self.prepare_plot` must have been called previous to this.
        
        Parameters:
        
        source_label (str): string corresponding to a label previously 
            called in `self.add_source`
        bokeh_model: any Bokeh model or glyph class
        tooltips: string or list of tuples (passed to Bokeh HoverTool)
        legend (str): name to assign to this layer in the plot legend
        start_hex (str): the color with which to start a color gradient
        end_hex (str): the color with which to end a color gradient
        mid_hex (str): the color to use for zero in a color gradient if 
            the basis of that gradient contains both positive and negative 
            values
        color_transform (callable): any function that can transform a 
            numpy array (log, log10, etc.)
        kwargs: options passed to the objected for `bokeh_model`
        
        This method allows two special kwargs: 'color' and 'alpha'. When 
        used with a bokeh model that has 'fill_color' and 'line_color' and 
        'fill_alpha' and 'line_alpha' properties, calling the special kwarg 
        will use the same value for both.
        
        """
    
        self._validate_workflow('add_layer')
        
        if self.plot is None:
            raise AssertionError('self.plot is null; call `self.prepare_plot`.')

        if bokeh_model not in bokeh_utils.MODELS:
            raise ValueError(
                'Valid values for `bokeh_model` are: {}'.format(
                    ', '.join([repr(x) for x in bokeh_utils.MODELS.keys()])
                )
            )
            
        bokeh_model = bokeh_utils.MODELS[bokeh_model]
        kwargs, hover_kwargs, new_fields, colorbar = bokeh_utils.prepare_properties(
            bokeh_model, kwargs, self.sources[source_label], bar_height=self.plot.frame_height,
            start_hex=start_hex, end_hex=end_hex, mid_hex=mid_hex,
            color_transform=color_transform,
        )

        if self.colorbar is None:
            self.colorbar = colorbar
                
        source = self.columndatasources[source_label]

        for k, v in new_fields.items():
            self.columndatasources[source_label].data[k] = v

        hover_object = None
        if bokeh_model == bokeh_utils.MODELS['MultiPolygons']:
            if type(self.plot) == GMapPlot:
                raise ValueError(
                        'The `MultiPolygon` glyph cannot yet be used with a Google Maps plot.'
                    )
            model_object = bokeh_model(xs='xsf', ys='ysf', name=source_label, **kwargs)

            if len(hover_kwargs) > 0:
                for k, v in hover_kwargs.items():
                    kwargs[k] = v
                hover_object = bokeh_model(xs='xsf', ys='ysf', name=source_label, **kwargs)

            if 'f' in self.remove_columns[source_label]:
                _ = self.remove_columns[source_label].pop(self.remove_columns[source_label].index('f'))

        else:

            model_object = bokeh_model(x='xsp', y='ysp', name=source_label, **kwargs)

            if len(hover_kwargs) > 0:
                for k, v in hover_kwargs:
                    kwargs[k] = v
                hover_object = bokeh_model(xs='xsf', ys='ysf', name=source_label, **kwargs)

            if 'p' in self.remove_columns[source_label]:
                _ = self.remove_columns[source_label].pop(self.remove_columns[source_label].index('p'))

        if source_label in self.views:
            if len(hover_kwargs) > 0:
                rend = self.plot.add_glyph(
                    source, model_object, hover_glyph=hover_object, view=self.views[source_label]
                )
            else:
                rend = self.plot.add_glyph(source, model_object, view=self.views[source_label])
        else:
            if len(hover_kwargs) > 0:
                rend = self.plot.add_glyph(source, model_object, hover_glyph=hover_object)
            else:
                rend = self.plot.add_glyph(source, model_object)

        if legend is not None:
            li = LegendItem(label=legend, renderers=[rend])
            self.legend.items.append(li)
    
        if tooltips is not None:
            if tooltips == 'all':
                tooltips = [
                    (k, '@{}'.format(k)) for k in source.data.keys()
                    if k not in ('xsf', 'ysf', 'xsp', 'ysp')
                ]
            elif tooltips == 'point':
                tooltips = [
                    (k, '@{}'.format(k)) for k in source.data.keys()
                    if k not in ('xsf', 'ysf', 'xsp', 'ysp', 'raw_data')
                ]
            elif tooltips == 'raw_data':
                tooltips = [
                    (k, '@{}'.format(k)) for k in source.data.keys()
                    if k not in ('xsf', 'ysf', 'xsp', 'ysp', 'x_coord_point', 'y_coord_point')
                ]
            elif tooltips == 'meta':
                tooltips = [
                    (k, '@{}'.format(k)) for k in source.data.keys()
                    if k not in ('xsf', 'ysf', 'xsp', 'ysp', 'x_coord_point', 'y_coord_point', 'raw_data')
                ]

        self.plot.add_tools(HoverTool(tooltips=tooltips, renderers=[rend]))

        if click_for_map is not None:
            taptool = self.plot.select(type=TapTool)
            if click_for_map == 'google':
                url = 'https://maps.google.com/maps?q=@y_coord_point,@x_coord_point'
                taptool.callback = OpenURL(url=url)
            elif click_for_map == 'bing':
                url = 'https://bing.com/maps/default.aspx?sp=point.'
                url += '@{y_coord_point}_@{x_coord_point}_Selected point&style=r'
                taptool.callback = OpenURL(url=url)
            else:
                raise NotImplementedError('Value for `click_for_map` must be "bing", "google" or None.')

        self.validation['add_layer'] = True
        
        return self

    def add_path(self, source_label, links=None, edge_type='curved', tooltips=None, legend=None, **kwargs):
        """
        Connect all points in the datasource in a path (to show order).
        `self.prepare_plot` must have been called previous to this.
        
        Parameters:
        
        source_label (str): string corresponding to a label previously 
            called in `self.add_source`
        order_col: the column of a data source specifying the order of the records
        tooltips: string or list of tuples (passed to Bokeh HoverTool)
        legend (str): name to assign to this layer in the plot legend
        kwargs: options passed to the objected for `bokeh_model`
        
        This method allows two special kwargs: 'color' and 'alpha'. When 
            used with a bokeh model that has 'fill_color' and 'line_color' 
            and 'fill_alpha' and 'line_alpha' properties, calling the special 
            kwarg will use the same value for both.
        
        """
    
        self._validate_workflow('add_path')
        
        source = self.sources[source_label].copy()
        
        suffix = '_transform' if type(self.plot) != GMapPlot else ''
        x_point_label = 'x_coord_point{}'.format(suffix)
        y_point_label = 'y_coord_point{}'.format(suffix)

        if all(isinstance(x, (int, float)) for x in source[links].tolist()):
            x1 = source.sort_values(links)[x_point_label].values
            y1 = source.sort_values(links)[y_point_label].values
            x2 = x1[1:]
            y2 = y1[1:]
            x1 = x1[:-1]
            y1 = y1[:-1]
            x3 = (x1 + x2) / 2
            y3 = (y1 + y2) / 2
            xc = x3 + abs(y3-y2)
            yc = y3 + abs(x3-x2)

            new_source = {'x1': x1, 'x2': x2, 'xc': xc, 'y1': y1, 'y2': y2, 'yc': yc}

            for c in source.columns:
                if (c not in self.omit_columns) and c not in new_source:
                    new_source[c] = source[c].values[:-1] 
        elif all(isinstance(x, (list, tuple, set)) for x in source[links].tolist()):
            if 'uid' not in source.columns:
                raise ValueError('Source must contain column `uid` when links is a list of iterables.')
                
            nodes = source['uid'].tolist()
            edges = source[links].tolist()
            node_x = source.set_index('uid')[x_point_label].to_dict()
            node_y = source.set_index('uid')[y_point_label].to_dict()

            a_vals, x1, x2, y1, y2 = zip(*[
                (a, node_x[a], node_x[b], node_y[a], node_y[b]) 
                for a, bs in zip(nodes, edges) for b in bs
            ])
            x1, x2, y1, y2 = array(x1), array(x2), array(y1), array(y2)    
            x3 = (x1 + x2) / 2
            y3 = (y1 + y2) / 2
            xc = x3 + abs(y3-y2)
            yc = y3 + abs(x3-x2)

            new_source = {'x1': x1, 'x2': x2, 'xc': xc, 'y1': y1, 'y2': y2, 'yc': yc}
            
            for c in source.columns:
                if (c not in self.omit_columns) and c not in new_source:
                    col_dict = source.set_index('uid', drop=False)[c].to_dict()
                    new_source[c] = [col_dict[v] for v in a_vals]
        else:
            raise ValueError(
                'Values of `links` field must be numeric or a list, set, or tuple of values from the `uid` field.'
            )
        
        if 'color' in kwargs.keys():
            color = kwargs.pop('color')
            for v in Quadratic.dataspecs():
                if 'color' in v:
                    kwargs[v] = color

        if 'alpha' in kwargs.keys():
            alpha = kwargs.pop('alpha')
            for v in Quadratic.dataspecs():
                if 'alpha' in v:
                    kwargs[v] = alpha
                
        if edge_type == 'curved':
            model_object = Quadratic(
                x0='x1', y0='y1', x1="x2", y1="y2", cx="xc", cy="yc", name=source_label, **kwargs
            )
        elif edge_type == 'straight':
            model_object = Segment(
                x0='x1', y0='y1', x1="x2", y1="y2", name=source_label, **kwargs
            )
        else:
            raise ValueError('Keyword `edge_type` must be either "curved" or "straight".')

        source = ColumnDataSource(new_source, name=source_label)
        rend = self.plot.add_glyph(source, model_object)
        
        if legend is not None:
            li = LegendItem(label=legend, renderers=[rend])
            self.legend.items.append(li)
    
        if tooltips is not None:
            self.plot.add_tools(HoverTool(tooltips=tooltips, renderers=[rend]))
            
        return self

    def add_data_table(self, source_label, columns='all', **kwargs):

        self._validate_workflow('add_data_table')

        source = self.columndatasources[source_label]

        if isinstance(columns, (list, tuple)):
            columns = {
                k: max([len(str(val)) for val in v] + [len(k)]) for k, v in source.data.items()
                if k in columns
            }
        else:
            if columns == 'all':
                omit_cols = ('xsf', 'ysf', 'xsp', 'ysp')
            elif columns == 'point':
                omit_cols = ('xsf', 'ysf', 'xsp', 'ysp', 'raw_data')
            elif columns == 'raw_data':
                omit_cols = ('xsf', 'ysf', 'xsp', 'ysp', 'x_coord_point', 'y_coord_point')
            elif columns == 'meta':
                omit_cols = ('xsf', 'ysf', 'xsp', 'ysp', 'x_coord_point', 'y_coord_point', 'raw_data')
            else:
                omit_cols = list()

            columns = {
                k: max([len(str(val)) for val in v] + [len(k)]) for k, v in source.data.items()
                if k not in omit_cols
            }

        default_kw = {
            'editable': False,
            'index_position': None,
            'reorderable': True,
            'scroll_to_selection': True,
            'selectable': 'checkbox',
            'sortable': True,
            'fit_columns': False
        }

        for k, v in default_kw.items():
            if k not in kwargs.keys():
                kwargs[k] = v

        data_table = DataTable(
            source=source,
            columns=[TableColumn(field=k, title=k, width=v * 8) for k, v in columns.items()],
            **kwargs
        )

        self.data_tables.append(data_table)

        return self
            
    def render_plot(
        self, display_type='object', directory=None, legend_position='below', 
        legend_orientation='horizontal', widget_position='left'
    ):
        """
        Pull everything together into a plot ready for display.
        
        Parameters:
        
        display_plot (str): either 'object', 'notebook', or an 
            arbitrary string. If 'object', it returns the plot object. 
            If 'notebook', the plot is displayed in the notebok. 
            If an arbitrary string that does not match one of the other pptions, 
            the plot is saved to '{display_plot}.html' in the current working 
            directory if `directory` is None, or in `directory` if not None.
            
        legend_position (str): 'below', 'above', 'left', or 'right'
        legend_orientation (str): 'horizontal' or 'vertical'
        widget_position (str): 'below', 'above', 'left', or 'right'
        
        """
        
        self._validate_workflow('render_plot')

        for k in self.columndatasources.keys():
            for suffix in self.remove_columns[k]:
                self.columndatasources[k].data.pop('xs' + suffix)
                self.columndatasources[k].data.pop('ys' + suffix)
        
        if len(self.legend.items) > 0:
            self.legend.orientation = legend_orientation
            self.plot.add_layout(self.legend, legend_position)

        self.plot.toolbar.autohide = self.autohide

        if self.colorbar is not None:
            self.plot = Row(children=[self.plot, self.colorbar])

        if len(self.widgets) > 0:

            if 'animation' in self.widgets.keys():
                animate = True
                button = self.widgets.pop('animation')
                button = button['widget']
            else:
                animate = False
                button = None

            widget_list = [d['widget'] for d in self.widgets.values()]

            if animate:
                if (len(widget_list) > 1) or not isinstance(widget_list[0], bokeh_utils.Slider):
                    raise NotImplementedError(
                        'Animations are currently only implented for plots that have only a Slider widget.'
                    )
                widget_list = [button] + widget_list

            if widget_position not in ('left', 'right', 'above', 'below'):
                raise ValueError("Valid widget positions are 'left', 'right', 'above', 'below'.")
            if widget_position == 'left':
                self.plot = Row(children=[WidgetBox(children=widget_list), self.plot])
            if widget_position == 'right':
                self.plot = Row(children=[self.plot, WidgetBox(children=widget_list)])
            if widget_position == 'above':
                self.plot = Column(children=[WidgetBox(children=widget_list), self.plot])
            if widget_position == 'below':
                self.plot = Column(children=[self.plot, WidgetBox(children=widget_list)])

        if len(self.data_tables) > 0:
            self.plot = Column(children=[self.plot] + self.data_tables)

        self.validation['render_plot'] = True
        
        if display_type == 'notebook':
            output_notebook(CDN, hide_banner=True, load_timeout=60000)
            show(self.plot)
        elif display_type == 'object':
            return self.plot
        else:
            if directory is not None:
                display_type = path.join(directory, display_type)
            save(self.plot, filename='{}.html'.format(display_type), resources=CDN, title=display_type)
            return display_type



from bokeh.colors import named
from numpy import array, repeat, argmax, linspace

NAMED_COLORS = {color: getattr(named, color).to_hex() for color in named.__all__}


def minmax_scale(x, out_range=(-1, 1)):
    if type(x) != array:
        x = array(x)
        
    domain = min(x), max(x)
    y = (x - (domain[1] + domain[0]) / 2) / (domain[1] - domain[0])
    
    return y * (out_range[1] - out_range[0]) + (out_range[1] + out_range[0]) / 2


def hex_to_rgb(hexcode):
    """'#FFFFFF' -> [255,255,255] """
    # Pass 16 to the integer function for change of base
    return [int(hexcode[i:i+2], 16) for i in range(1, 6, 2)]


def rgb_to_hex(rgb):
    """ [255,255,255] -> '#FFFFFF' """
    # Components need to be integers for hex to make sense
    rgb = [int(x) for x in rgb]
    return "#"+"".join(["0{0:x}".format(v) if v < 16 else "{0:x}".format(v) for v in rgb])


def check_color(color):
    if color is None:
        return True
    if color in NAMED_COLORS.keys():
        return True
    else:
        try:
            _ = hex_to_rgb(color)
            return True
        except (ValueError, IndexError, TypeError):
            return False
        

def check_numeric(val):
    if isinstance(val, (int, float)):
        return True
    elif isinstance(val, (list, tuple, set)):
        return all(isinstance(x, (int, float)) for x in val)
    else:
        return False

    
def color_gradient(vals, start_hex, end_hex, mid_hex='#ffffff', trans=None):
    """
    Returns a gradient list of colors between two hex colors (adapted 
    from https://bsou.io/posts/color-gradients-with-python). Gradient 
    is spaced linearly with the numerical values of (optionally transformed) 
    `vals`. If `vals` include both positive negative numbers, gradient goes 
    from `start_hex` to `mid_hex` (default of white at zero), to `end_hex`.
    Values for start_hex and end_hex should be the full six-digit color string, 
    including the number sign ("#FFFFFF").
    
    """
    
    # convert vals to array
    sarr = array(vals).astype('float64')
    has_neg = min(vals) < 0
    has_pos = max(vals) > 0
    
    # colors in RGB form
    s = hex_to_rgb(start_hex)
    f = hex_to_rgb(end_hex)
    m = hex_to_rgb(mid_hex)

    if not all([has_neg, has_pos]):
        if trans is not None:
            sarr = trans(minmax_scale(sarr, out_range=(1e-10, 10)))
        sarr = minmax_scale(sarr, out_range=(0, 1))
        
        return [rgb_to_hex(rgb) for rgb in zip(*[(s[j] + sarr * (f[j] - s[j])) for j in range(3)])]
    else:
        sarr_neg = sarr[sarr <= 0]
        sarr_pos = sarr[sarr >= 0]
        if trans is not None:
            sarr_pos = trans(minmax_scale(sarr_pos, out_range=(1e-10, 10)))
            sarr_neg = trans(minmax_scale(sarr_neg, out_range=(1e-10, 10)))

        sarr_pos = minmax_scale(sarr_pos, out_range=(0, 1))
        sarr_neg = minmax_scale(sarr_neg, out_range=(0, 1))
        
        sarr_new = repeat(mid_hex, sarr.shape[0])
        sarr_new[sarr >= 0] = [
            rgb_to_hex(rgb) 
            for rgb in zip(*[(m[j] + sarr_pos * (f[j] - m[j])) for j in range(3)])
        ]
        sarr_new[sarr <= 0] = [
            rgb_to_hex(rgb) 
            for rgb in zip(*[(s[j] + sarr_neg * (m[j] - s[j])) for j in range(3)])
        ]

        return sarr_new


def order_records(x_coords, y_coords):
    n_records = len(x_coords)
    record_inds = list(range(n_records))
    a = array(list(zip(x_coords, y_coords)))

    inds = array([True] * n_records)
    current_ind = 0
    ind_order = list()

    while any(inds):
        inds[current_ind] = False
        ind_order.append(current_ind)
        dists = ((a[current_ind, :] - a) ** 2).sum(axis=1) ** 0.5
        dists[~inds] = -1
        current_ind = argmax(dists)

    order_dict = dict(zip([record_inds[ind] for ind in ind_order], record_inds))
    score = [order_dict[ind] for ind in record_inds]
    return score


def _v(m1, m2, hue):
    hue = hue % 1.0
    if hue < (1 / 6):
        return m1 + (m2 - m1) * hue * 6.0
    if hue < 0.5:
        return m2
    if hue < (2 / 3):
        return m1 + (m2 - m1) * ((2 / 3) - hue) * 6.0
    return m1


def hls_to_rgb(h, l, s):
    """
    Adapted from https://github.com/python/cpython/blob/master/Lib/colorsys.py
    """
    if s == 0.0:
        return l, l, l
    if l <= 0.5:
        m2 = l * (1.0 + s)
    else:
        m2 = l + s - (l * s)
    m1 = 2.0 * l - m2
    r, g, b = _v(m1, m2, h + (1.0 / 3.0)), _v(m1, m2, h), _v(m1, m2, h - (1.0 / 3.0))
    return int(r * 255.0), int(g * 255.0), int(b * 255.0)


def rgb_to_hex_sep(r, g, b):
    return "#{0:02x}{1:02x}{2:02x}".format(r, g, b)


def hls_palette(n_colors=6, h=.01, l=.6, s=.65):
    """
    Get a set of evenly spaced colors in HLS hue space.
    h, l, and s should be between 0 and 1

    Adapted from https://github.com/mwaskom/seaborn/blob/master/seaborn/palettes.py

    Parameters
    ----------
    n_colors : int
        number of colors in the palette
    h : float
        first hue
    l : float
        lightness
    s : float
        saturation

    Returns
    -------
    palette : List-like object of colors as hex color strings.

    """

    hues = linspace(0, 1, int(n_colors) + 1)[:-1]
    hues += h
    hues %= 1
    hues -= hues.astype(int)
    palette = [rgb_to_hex_sep(*hls_to_rgb(h_i, l, s)) for h_i in hues]
    return palette


def assign_colors(color_arr, start=None, end=None, mid='#ffffff', trans=None, categorical_palette=None):

    if categorical_palette is None:
        categorical_palette = None
    
    # continuous palette
    if check_numeric(color_arr):
        if (start is not None) and (end is not None):
            return color_gradient(color_arr, start_hex=start, end_hex=end, mid_hex=mid, trans=trans)
        else:
            raise ValueError('Values for `start` and `end` must be supplied for numeric arrays.')

    # categorical palette
    value_set = set(color_arr)
    n_colors = len(value_set)
    palette_length = len(categorical_palette)

    if n_colors <= palette_length:
        color_dict = dict(zip(value_set, categorical_palette[:n_colors]))
    else:
        val_count_dict = dict()
        for val in color_arr:
            if val in val_count_dict.keys():
                val_count_dict[val] += 1
            else:
                val_count_dict[val] = 1
        color_dict = dict()
        n = 0
        for val, num in sorted(val_count_dict.items(), reverse=True):
            color_dict[val] = categorical_palette[n]
            if n < (palette_length - 1):
                n += 1

    return [color_dict[v] for v in color_arr]



from bokeh.models import CheckboxGroup, CheckboxButtonGroup, RangeSlider, Slider, Dropdown, RadioButtonGroup, Button
from bokeh.models import CustomJS
from bokeh.events import ButtonClick
from bokeh.models import markers, WMTSTileSource
from bokeh.models.glyphs import MultiPolygons, Text
from bokeh.models import Plot, Rect, ColumnDataSource

from .color_utils import check_color, check_numeric, color_gradient, hls_palette, order_records

# non-Google-Map map tile sources
# adapted from https://github.com/holoviz/holoviews/blob/master/holoviews/element/tiles.py


def get_tile_source(name):
    tiles = {
        # CartoDB basemaps
        'carto_dark': 'https://cartodb-basemaps-4.global.ssl.fastly.net/dark_all/{Z}/{X}/{Y}.png',
        'carto_eco': 'https://3.api.cartocdn.com/base-eco/{Z}/{X}/{Y}.png',
        'carto_light': 'https://cartodb-basemaps-4.global.ssl.fastly.net/light_all/{Z}/{X}/{Y}.png',
        'carto_midnight': 'https://3.api.cartocdn.com/base-midnight/{Z}/{X}/{Y}.png',

        # Stamen basemaps
        'stamen_terrain': 'https://tile.stamen.com/terrain/{Z}/{X}/{Y}.png',
        'stamen_terrain_retina':  'https://tile.stamen.com/terrain/{Z}/{X}/{Y}@2x.png',
        'stame_watercolor': 'https://tile.stamen.com/watercolor/{Z}/{X}/{Y}.jpg',
        'stamen_toner': 'https://tile.stamen.com/toner/{Z}/{X}/{Y}.png',
        'stamen_toner_background': 'https://tile.stamen.com/toner-background/{Z}/{X}/{Y}.png',
        'stamen_toner_labels': 'https://tile.stamen.com/toner-labels/{Z}/{X}/{Y}.png',

        # Esri maps (see https://server.arcgisonline.com/arcgis/rest/services for the full list)
        'esri_imagery': 'https://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{Z}/{Y}/{X}.jpg',
        'esri_natgeo': 'https://server.arcgisonline.com/ArcGIS/rest/services/NatGeo_World_Map/MapServer/tile/{Z}/{Y}/{X}',
        'esri_usatopo': 'https://server.arcgisonline.com/ArcGIS/rest/services/USA_Topo_Maps/MapServer/tile/{Z}/{Y}/{X}',
        'esri_terrain': 'https://server.arcgisonline.com/ArcGIS/rest/services/World_Terrain_Base/MapServer/tile/{Z}/{Y}/{X}',
        'esri_reference': 'https://server.arcgisonline.com/ArcGIS/rest/services/Reference/World_Reference_Overlay/MapServer/tile/{Z}/{Y}/{X}',

        # Miscellaneous
        'osm': 'https://c.tile.openstreetmap.org/{Z}/{X}/{Y}.png',
        'osm_bw': 'https://tiles.wmflabs.org/bw-mapnik/{z}/{x}/{y}.png',
        'wikipedia': 'https://maps.wikimedia.org/osm-intl/{Z}/{X}/{Y}@2x.png'
    }

    if name is None:
        return list(tiles.keys())

    if name not in tiles:
        return None

    return WMTSTileSource(url=tiles[name])


# all supported models
MODELS = {k: getattr(markers, k) for k in markers.__all__}
MODELS['MultiPolygons'] = MultiPolygons
MODELS['Text'] = Text

MODELS_REVERSE = {v: k for k, v in MODELS.items()}

# all supported widgets
WIDGETS = {
    'Dropdown': Dropdown, 
    'RangeSlider': RangeSlider, 
    'CheckboxGroup': CheckboxGroup, 
    'CheckboxButtonGroup': CheckboxButtonGroup,
    'Slider': Slider,
    'RadioButtonGroup': RadioButtonGroup
}

DEFAULT_KWARGS = {
    'CheckboxGroup': {
        'labels': 'auto',
        'active': 'auto',
        'width': 90
    },
    'RangeSlider': {
        'start': 'auto',
        'end': 'auto',
        'step': 1,
        'value': 'auto',
        'width': 50,
        'show_value': True,
        'orientation':
        'vertical',
        'tooltips': False,
        'format': '0'
    },
    'Dropdown': {
        'menu': 'auto',
        'value': 'auto',
        'width': 100,
        'label': 'geohashes'
    },
    'Slider': {
        'start': 'auto',
        'end': 'auto',
        'step': 1,
        'value': 'auto',
        'width': 50,
        'show_value': True,
        'orientation': 'vertical',
        'tooltips': False,
        'format': '0'
    },
    'CheckboxButtonGroup': {
        'labels': 'auto',
        'active': 'auto'
    },
    'RadioButtonGroup': {
        'labels': 'auto',
        'active': 'auto'
    }
}

# filter boilerplate
FILTERS = {
    'Dropdown': '''
        var indices = [];
        for (var i = 0; i <= source.data[reference].length; i++){
            if (source.data[reference][i] == widget.value) {
                indices.push(i)
            }
        }
        return indices;
    ''',
    'Slider': '''
        var indices = [];
        if (widget.step % 1 === 0){
            var value = Math.trunc(widget.value)
        } else {
            var value = widget.value
        }
        
        for (var i = 0; i <= source.data[reference].length; i++){
            if (source.data[reference][i] == widget.value) {
                indices.push(i)
            }
        }
        return indices;
    ''',
    'RangeSlider': '''
        var indices = [];
        if (widget.step % 1 === 0){
            var lower = Math.trunc(widget.value[0])
            var upper = Math.trunc(widget.value[1])
        } else {
            var lower = widget.value[0]
            var upper = widget.value[1]        
        }
        for (var i = 0; i <= source.data[reference].length; i++){
            var value = source.data[reference][i]
            if (value >= lower) {
                if (value <= upper) {
                    indices.push(i)
                }
            }
        }
        return indices;
    ''',
    'CheckboxGroup': '''
        var indices = [];
        var results = [];

        for (var i = 0; i < widget.active.length; i++)
            results.push(widget.labels[widget.active[i]]);

        for (var i = 0; i < source.data[reference].length; i++){
            var value_string = source.data[reference][i].toString()
            if (results.includes(value_string)) {
                indices.push(i)
            }
        }
        return indices;
    ''',
    'CheckboxButtonGroup': '''
        var indices = [];
        var results = [];

        for (var i = 0; i < widget.active.length; i++)
            results.push(widget.labels[widget.active[i]]);

        for (var i = 0; i < source.data[reference].length; i++){
            var value_string = source.data[reference][i].toString()
            if (results.includes(value_string)) {
                indices.push(i)
            }
        }
        return indices;
    ''',
    'RadioButtonGroup': '''
        var indices = [];
        var results = [];

        var result = widget.labels[widget.active];

        for (var i = 0; i < source.data[reference].length; i++){
            var value_string = source.data[reference][i].toString()
            if (result == value_string) {
                indices.push(i)
            }
        }
        return indices;
    '''
}


def auto_advance(slider, ms_delay=500):
    return CustomJS(args=dict(slider=slider), code="""

            function sleep(ms) {
                return new Promise(resolve => setTimeout(resolve, ms));
            }

            async function autoAdvance() {

                slider.value = slider.start;

                while (slider.value < slider.end) {
                    await sleep(%d)
                    slider.value = slider.value + slider.step
                };
            }

            autoAdvance();
        """ % ms_delay
                    )


def auto_widget_kwarg(widget_type, kwarg, reference_array):
    """
    For a particular widget type, keyword argument, and array
    of values, set reasonable defaults.
    
    """
    
    if widget_type in ("CheckboxGroup", "CheckboxButtonGroup"):
        reference_set = list(set(reference_array))
        if kwarg == 'labels':
            return reference_set
        if kwarg == 'active':
            return [x for x in range(len(reference_set))]
        raise ValueError(
            'The only auto-populating kwargs for {} are `labels` and `active`.'.format(widget_type)
        )
    if widget_type == "RangeSlider":
        if kwarg == 'start':
            return min(reference_array)
        if kwarg == 'end':
            return max(reference_array)
        if kwarg == 'value':
            return min(reference_array), max(reference_array)
        raise ValueError(
            'The only auto-populating kwargs for {} are `start`, `end` and `value`.'.format(widget_type)
        ) 
    if widget_type == "Slider":
        if kwarg == 'start':
            return min(reference_array)
        if kwarg == 'end':
            return max(reference_array)
        if kwarg == 'value':
            return min(reference_array)
        raise ValueError(
            'The only auto-populating kwargs for {} are `start`, `end` and `value`.'.format(widget_type)
        )
    if widget_type == "Dropdown":
        reference_set = list(set(reference_array))
        if kwarg == 'menu':
            return reference_set
        if kwarg == 'value':
            return reference_set[0]
        raise ValueError(
            'The only auto-populating kwargs for {} are `menu` and `value`.'.format(widget_type)
        )    
    if widget_type in ("RadioButtonGroup", ):
        reference_set = list(set(reference_array))
        if kwarg == 'labels':
            return reference_set
        if kwarg == 'active':
            return 0
        raise ValueError(
            'The only auto-populating kwargs for {} are `labels` and `active`.'.format(widget_type)
        )
    
    raise NotImplementedError('Widget type `{}` not yet implemented.'.format(widget_type))

    
def prepare_properties(
    bokeh_model, kwargs, source_df, bar_height,
    start_hex='#ff0000', end_hex='#0000ff', mid_hex='#ffffff', color_transform=None
):
    """
    For a particular Bokeh model and accompanying keyword arguents,
    automatically set color and alpha values.
    
    """

    colorbar = None

    hover_kwargs = dict()
    for k, v in kwargs.items():
        if k.startswith('hover_'):
            hover_kwargs[k.replace('hover_', '')] = v

    for k in hover_kwargs.keys():
        _ = kwargs.pop('hover_' + k)

    if 'color' in kwargs.keys():
        color = kwargs.pop('color')
        for v in bokeh_model.dataspecs():
            if 'color' in v:
                kwargs[v] = color

    color_keys = [key for key in bokeh_model.dataspecs() if ('color' in key) and (key in kwargs.keys())]

    new_fields = dict()
    for key in color_keys:
        color_val = kwargs[key]

        if color_val is None:
            continue

        if isinstance(color_val, str):
            if check_color(color_val):
                continue
            else:
                color_arr = source_df[color_val].tolist()
                in_datasource = True
        else:
            color_arr = color_val
            in_datasource = False

        if not all(check_color(c) for c in color_arr):

            if check_numeric(color_arr):
                if (start_hex is not None) and (end_hex is not None):
                    color_new = color_gradient(
                        color_arr,
                        start_hex=start_hex, end_hex=end_hex, mid_hex=mid_hex,
                        trans=color_transform
                    )

                    colorbar = make_colorbar(color_arr, color_new, bar_height)
                else:
                    raise ValueError('Values for `start_hex` and `end_hex` must be supplied for numeric arrays.')
            else:
                n_colors = len(set(color_arr))
                color_df = source_df.groupby(color_arr)[['x_coord_point', 'y_coord_point']].mean()
                score = order_records(color_df['x_coord_point'].tolist(), source_df['y_coord_point'].tolist())
                palette = hls_palette(n_colors, h=0.5, l=0.5, s=1.0)
                color_dict = dict(zip(color_df.index.tolist(), [palette[ind] for ind in score]))
                color_new = [color_dict[ind] for ind in color_arr]

            if in_datasource:
                new_val = '{}_autocolor'.format(color_val)
                new_fields[new_val] = color_new
                kwargs[key] = new_val
            else:
                kwargs[key] = color_new

    if 'alpha' in kwargs.keys():
        alpha = kwargs.pop('alpha')
        for v in bokeh_model.dataspecs():
            if 'alpha' in v:
                kwargs[v] = alpha
                    
    return kwargs, hover_kwargs, new_fields, colorbar


def make_colorbar(values, colors, bar_height):
    val_list = list()
    seen = set()
    for val, color in zip(values, colors):
        if color not in seen:
            val_list.append((val, color))
            seen.add(color)

    val_opts, color_opts = zip(*sorted(val_list))
    n_opts = len(val_opts)
    longest_val = len(max([str(v) for v in val_opts], key=len))
    opt_inds = [v for v in range(n_opts)]
    max_color, max_val, max_ind = color_opts[0], val_opts[0], opt_inds[0]
    min_color, min_val, min_ind = color_opts[-1], val_opts[-1], opt_inds[-1]

    colorbar = Plot(
        frame_width=10 * longest_val, frame_height=bar_height,
        title=None, min_border=0,
        toolbar_location=None, outline_line_color=None
    )
    colorbar_source = ColumnDataSource({'x': [0] * n_opts, 'y': opt_inds, 'vals': val_opts, 'colors': color_opts})
    glyph = Rect(x=1, y="y", width=1, height=1, fill_color="colors", line_color=None)
    text_source = ColumnDataSource({'x': [1, 1], 'y': [min_ind + 1, max_ind - 1], 'text': [str(min_val), str(max_val)]})
    text = Text(x=1, y='y', text='text', text_font_size='10pt', text_align='center', text_baseline='middle')
    colorbar.add_glyph(colorbar_source, glyph)
    colorbar.add_glyph(text_source, text)

    return colorbar



from setuptools import setup, find_packages

try:
    with open('README.md') as f:
        long_description = f.read()
except:
    long_description = ''
    print('Failed to load README.md as long_description')

setup(
    name='theto',
    version='0.2.1',
    description='Workflow automation for exploring location data',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Valassis Digital',
    url='https://github.com/Valassis-Digital-Media/theto',
    platforms=['Linux', 'Mac OSX', 'Windows'],
    packages=find_packages(),
    include_package_data=True,
    license='BSD 3-Clause',
    install_requires=[
        'bokeh>=1.0',
        'pandas',
        'numpy',
        'shapely',
        'pyproj'
    ],
)


from math import sin, pi, log, floor


def lat_rad(lat):
    """Convert a latitude to radians (for estimating zoom factor for Google Maps)."""
    sine = sin(lat * pi / 180.)
    rad_x2 = log((1 + sine) / (1 - sine)) / 2.
    return max(min(rad_x2, pi), -pi) / 2.

    
def zoom(map_px, world_px, fraction):
    """Calculate the zoom factor for Google Maps."""
    try:
        return floor(log(map_px / world_px / fraction) / log(2))
    except ZeroDivisionError:
        return None

    
def estimate_zoom(plot_width, x_bounds, y_bounds, height_max=600, zoom_max=21, world_dim=256):
    """
    Given a desired plot width and data source(s), estimate the best zoom factor for Google Maps.
    
    Parameters:
     
    plot_width (int): desired plot width, in pixels
    x_bounds (tuple): min and max of x axis
    y_bounds (tuple): min and max of y axis
    height_max (int): maximum height allowable for the plot
    zoom_max (int): maximum zoom factor (21 is the maximum Google Maps allows)
    world_dim (int): the number of dimensions fo the world map (256 is the Google default)
        
    Returns:
        
    Zoom factor (int): the zoom factor to be used in the call to Google Maps
    y_center (float): the central latitude for the plot
    x_center (float): the central longitude for the plot
    plot_height (int): the height to be used for the plot
        
    """

    xmin, xmax = x_bounds
    ymin, ymax = y_bounds
    y_range = abs(ymax - ymin)
    x_range = abs(xmax - xmin)
    if (x_range == 0) or (y_range == 0):
        plot_height = height_max
    else:
        plot_height = int((plot_width / x_range) * y_range)
    if plot_height < plot_width:
        plot_height = plot_width
    if plot_height > height_max:
        plot_height = height_max
    y_center = (ymin + ymax) / 2.0
    x_center = (xmin + xmax) / 2.0

    lat_fraction = (lat_rad(ymax) - lat_rad(ymin)) / pi
    lng_diff = xmax - xmin
    lng_fraction = ((lng_diff + 360) if (lng_diff < 0) else lng_diff) / 360.

    lat_zoom = zoom(plot_height, world_dim, lat_fraction)
    lng_zoom = zoom(plot_width, world_dim, lng_fraction)

    if (lat_zoom is None) and (lng_zoom is None):
        return zoom_max, y_center, x_center, plot_height
    elif lat_zoom is None:
        return int(min(lng_zoom, zoom_max)), y_center, x_center, plot_height
    elif lng_zoom is None:
        return int(min(lat_zoom, zoom_max)), y_center, x_center, plot_height
    else:
        return int(min(lat_zoom, lng_zoom, zoom_max)), y_center, x_center, plot_height
